//////////////////////////////////////////////////////////////////////////////
//
// Mesh.cpp, 12/08/01, Ilja Friedel
//
//
//////////////////////////////////////////////////////////////////////////////

#include "Mesh.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#define EXACT_DERIVATIVES
#define SHARP_BOUNDARY_CORNER
//#define EXPORT_IV_CONTROL_SPHERES

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
MeshPoint::~MeshPoint()
{

}

//////////////////////////////////////////////////////////////////////////////
// ok

MeshPoint::MeshPoint(FILE *f, double *P, int id)
{
  patches_in_ring= 0;
  load(f,P,id);
  //print(P);
}

//////////////////////////////////////////////////////////////////////////////
// load in Patrick and Ilja's format
int MeshPoint::load(FILE *f, double *P, int id)
{
  double x,y,z;
  int valence; // ***************************************
  assert(id>=0); // no virtual point
  patches_in_ring= 0;
  fscanf(f,"%d %d %d %le %le %le\n",&ID, &constraint, &valence, &x,&y,&z);
  //fscanf(f,"%d %le %le %le\n",&constraint, &x,&y,&z);
  //ID=id;
  //constraint=1;
  P[3*ID]  =x;//1.0*x;
  P[3*ID+1]=y;//5.0*y+z;
  P[3*ID+2]=z;//0.5*z;

  cout << x << "  " << y << "  " << z << endl;

  return(1);
}

//////////////////////////////////////////////////////////////////////////////
// save in Patrick and Ilja's format
int MeshPoint::save(FILE *f, double *P)
{
  int valence;
  valence=0;
  fprintf(f,"%d %d %d %e %e %e\n",ID,constraint,valence,X(P),Y(P),Z(P));

  return(1);
}
//////////////////////////////////////////////////////////////////////////////
// export point in IV format
int MeshPoint::export_IV(char *fname, double *P)
{
  FILE *f;

  if(ring_table)
    if(ring_table[0]->M->export_iv_sphere_size>0.0)
      {
	f=fopen(fname,"a+");
	assert(f);
	
	fprintf(f,"Separator \n");
	fprintf(f,"{\n");
	fprintf(f,"  Material\n");
	fprintf(f,"    {\n");

	if(constraint==0)
	  {
	    fprintf(f,"      diffuseColor 1 0 0.2\n");
	  }
	else
	  {
	    fprintf(f,"      diffuseColor 0 1 1\n");
	  }

	fprintf(f,"    }\n");
	fprintf(f,"  Translation\n");
	fprintf(f,"    {\n");
	fprintf(f,"      translation %e %e %e\n",X(P),Y(P),Z(P));
	fprintf(f,"    }\n");
	fprintf(f,"  Sphere\n");
	fprintf(f,"    { \n");
	fprintf(f,"      radius %1.4e\n",0.01*ring_table[0]->M->export_iv_sphere_size);
	fprintf(f,"    }\n");
	fprintf(f,"}\n");
	
	fclose(f);
      }
  return(1);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// ok
MeshPatch::~MeshPatch()
{
  if(points) delete[] points;
  points=NULL;
}

//////////////////////////////////////////////////////////////////////////////
// ok
MeshPatch::MeshPatch(Mesh *m, FILE *f)
{
  all_points_constrained=0; // standard asumption
  M=m;
  patch=NULL;
  boundary_type=0;
  load(f);//,pointtable);
}

//////////////////////////////////////////////////////////////////////////////
// copy index [into MeshPoint array]
MeshPatch::MeshPatch(int v, int *P)
{
  int i;

  all_points_constrained=0; // standard asumption
  patch=NULL;
  valence=v;
  boundary_type=0;
  points=new int[2*valence+8];
  for(i=0;i<2*valence+8;i++) points[i]=P[i]; 
}

//////////////////////////////////////////////////////////////////////////////
// load in Patrick and Ilja's file format
int MeshPatch::load(FILE *f)
{
  int i,tmp;
  
  patch=NULL;

  fscanf(f,"%d ",&valence);
  
  if(valence==0)
    {
      cout << "Warning: MeshPatch::load() ... Reached EOF."<<endl;
      return(0);
    }
  else
    {
      points=new int[2*valence+8];
      for(i=0;i<2*valence+8;i++)
	{
	  fscanf(f,"%d",&tmp);
	  cout << tmp << " ";
	  points[i]=tmp;
	}
      if(points[5]==-1 || points[6]==-1 || points[7]==-1 || points[8]==-1)
	{
	  assert(valence==4);
	  assert(points[ 5]==-1);
	  assert(points[ 6]==-1);
	  assert(points[ 7]==-1);
	  assert(points[ 8]==-1);
	  boundary_type=1;
	}
      if(points[9]==-1 || points[10]==-1 || points[11]==-1)
	{
	  assert(valence==4);
	  assert(points[ 5]==-1);
	  assert(points[ 6]==-1);
	  assert(points[ 7]==-1);
	  assert(points[ 8]==-1);
	  assert(points[ 9]==-1);
	  assert(points[10]==-1);
	  assert(points[11]==-1);    
 	  boundary_type=2;
	}
      cout << "BoundaryType="<<boundary_type<<endl;
    }
  cout << endl;
  return(1);  
}

//////////////////////////////////////////////////////////////////////////////
// save in Patrick and Ilja's file format 
int MeshPatch::save(FILE *f)
{
  int i;

  fprintf(f,"%d ",valence);
  for(i=0;i<2*valence+8;i++)
    {
      fprintf(f,"%d ",points[i]);
    }
  fprintf(f,"\n");

  return(1);
}

//////////////////////////////////////////////////////////////////////////////
// write patch to file in IV format
int MeshPatch::export_IV(char *fname, double *MP)
{
  ExtraordinaryPatch *P;
  FILE *f;
  double cr,cg,cb;
  int i;

  cr=0.5;
  cg=0.5;
  cb=0.5;

  //if(!patch) patch=create_Patch(P);
  //else update_Patch(P);
  //patch->export_IV(fname,cr,cg,cb);
  //delete EP;
  P=create_Patch(MP);

  if(M->export_iv_coloring_scheme==0) P->export_IV_phong(fname,
							 M->export_iv_M,
							 M->export_iv_N,
							 M->export_iv_reflection_map);
  if(M->export_iv_coloring_scheme==1) P->export_IV_principal_coloring(fname,
								      M->export_iv_coloring_range,
								      M->export_iv_M,
								      M->export_iv_N,
								      M->export_iv_reflection_map);
  if(M->export_iv_coloring_scheme==2) P->export_IV_nice_coloring(fname,
								 M->export_iv_coloring_range,
								 M->export_iv_M,
								 M->export_iv_N,
								 M->export_iv_reflection_map);

  if(M->export_iv_coloring_scheme==3 && !all_points_constrained) P->export_IV_nice_coloring(fname,
								 M->export_iv_coloring_range,
								 M->export_iv_M,
								 M->export_iv_N,
								 M->export_iv_reflection_map);
  //P->export_IV_mean_coloring(fname,50.0);
  //P->export_IV_gaussian_coloring(fname,10.0);
  //P->export_IV_principal_coloring(fname,10.0);

  if(M->export_iv_control_grid==1)
    {
      f=fopen(fname,"a+");
      assert(f);

      fprintf(f,"   Separator {\n");
      fprintf(f,"           Material {\n");
      fprintf(f,"                ambientColor   1 0 1 diffuseColor 1 0 1\n");
      fprintf(f,"           }\n");
      fprintf(f,"        Coordinate3 {\n");
      fprintf(f,"            point [\n");
      
      for(i=0;i<4;i++)
	{
	  fprintf(f,"%e %e %e",
		  P->control_points[0][i], 
		  P->control_points[1][i], 
		  P->control_points[2][i]);
	  
	  if(i<3) fprintf(f,", ");
	  else fprintf(f,"\n");
	}
      
      fprintf(f,"                  ]\n");
      fprintf(f,"        }\n");
      fprintf(f,"        IndexedLineSet {\n");
      fprintf(f,"            coordIndex [ 0, 1, 2, 3, -1\n");
      fprintf(f,"                       ]\n");
      fprintf(f,"        }");
      fprintf(f,"    }\n");
      fclose(f);
    }
  else if(M->export_iv_control_grid==2 && !all_points_constrained)
    {
      f=fopen(fname,"a+");
      assert(f);

      fprintf(f,"   Separator {\n");
      fprintf(f,"           Material {\n");
      fprintf(f,"                ambientColor   1 0 1 diffuseColor 1 0 1\n");
      fprintf(f,"           }\n");
      fprintf(f,"        Coordinate3 {\n");
      fprintf(f,"            point [\n");
      
      for(i=0;i<4;i++)
	{
	  fprintf(f,"%e %e %e",
		  P->control_points[0][i], 
		  P->control_points[1][i], 
		  P->control_points[2][i]);
	  
	  if(i<3) fprintf(f,", ");
	  else fprintf(f,"\n");
	}
      
      fprintf(f,"                  ]\n");
      fprintf(f,"        }\n");
      fprintf(f,"        IndexedLineSet {\n");
      fprintf(f,"            coordIndex [ 0, 1, 2, 3, 0, -1\n");
      fprintf(f,"                       ]\n");
      fprintf(f,"        }");
      fprintf(f,"    }\n");
      fclose(f);
    }
  else if(M->export_iv_control_grid==3 && all_points_constrained)
    {
      f=fopen(fname,"a+");
      assert(f);

      fprintf(f,"#Inventor V2.0 ascii\n");
      fprintf(f,"   Separator {\n");
      fprintf(f,"           Material {\n");
      fprintf(f,"                ambientColor   1 0 1 diffuseColor 1 0 1\n");
      fprintf(f,"           }\n");
      fprintf(f,"        Coordinate3 {\n");
      fprintf(f,"            point [\n");
      
      for(i=0;i<4;i++)
	{
	  fprintf(f,"%e %e %e",
		  P->control_points[0][i], 
		  P->control_points[1][i], 
		  P->control_points[2][i]);
	  
	  if(i<3) fprintf(f,", ");
	  else fprintf(f,"\n");
	}
      
      fprintf(f,"                  ]\n");
      fprintf(f,"        }\n");
      fprintf(f,"        IndexedLineSet {\n");
      fprintf(f,"            coordIndex [ 0, 1, 2, 3, 0, -1\n");
      fprintf(f,"                       ]\n");
      fprintf(f,"        }");
      fprintf(f,"    }\n");
      fclose(f);
    }

  delete P;
  return(1);
}

//////////////////////////////////////////////////////////////////////////////
// create a patch with the coordinates given in MP, valence given by this object
FirstOrderDataDependentPatch* MeshPatch::create_Patch(double *MP)
{
  FirstOrderDataDependentPatch *EP;
  double *data[3];
  double tmp1[2*MAX_VALENCE+8];
  double tmp2[2*MAX_VALENCE+8];
  double tmp3[2*MAX_VALENCE+8];
  int i;
      
  data[0]=tmp1;
  data[1]=tmp2;
  data[2]=tmp3;
  
  if(boundary_type==0)
    {
      for(i=0;i<2*valence+8;i++)
	{
	  assert(points[i]!=-1);
	  data[0][i]=MP[3*points[i]];
	  data[1][i]=MP[3*points[i]+1];
	  data[2][i]=MP[3*points[i]+2];
	}
    }
  else if(boundary_type==1)
    {
      assert(valence==4);

      //cout << "BT1" << endl;
      for(i=0;i<5;i++)
	{
	  assert(points[i]!=-1);
	  data[0][i]=MP[3*points[i]  ];
	  data[1][i]=MP[3*points[i]+1];
	  data[2][i]=MP[3*points[i]+2];
	}

      // mirroring of virtual points (across the boundary)
      data[0][5] = 2.0*MP[3*points[4]  ] - MP[3*points[2*valence+7]  ];
      data[1][5] = 2.0*MP[3*points[4]+1] - MP[3*points[2*valence+7]+1];
      data[2][5] = 2.0*MP[3*points[4]+2] - MP[3*points[2*valence+7]+2];
      
      data[0][6] = 2.0*MP[3*points[1]  ] - MP[3*points[ 0]  ];
      data[1][6] = 2.0*MP[3*points[1]+1] - MP[3*points[ 0]+1];
      data[2][6] = 2.0*MP[3*points[1]+2] - MP[3*points[ 0]+2];
      
      data[0][7] = 2.0*MP[3*points[2]  ] - MP[3*points[ 3]  ];
      data[1][7] = 2.0*MP[3*points[2]+1] - MP[3*points[ 3]+1];
      data[2][7] = 2.0*MP[3*points[2]+2] - MP[3*points[ 3]+2];
      
      data[0][8] = 2.0*MP[3*points[9]  ] - MP[3*points[10]  ];
      data[1][8] = 2.0*MP[3*points[9]+1] - MP[3*points[10]+1];
      data[2][8] = 2.0*MP[3*points[9]+2] - MP[3*points[10]+2];
      
      for(i=9;i<2*valence+8;i++)
	{
	  assert(points[i]!=-1);
	  data[0][i]=MP[3*points[i]  ];
	  data[1][i]=MP[3*points[i]+1];
	  data[2][i]=MP[3*points[i]+2];
	}
    }
  else if(boundary_type==2)
    {
      for(i=0;i<5;i++)
	{
	  assert(points[i]!=-1);
	  data[0][i]=MP[3*points[i]  ];
	  data[1][i]=MP[3*points[i]+1];
	  data[2][i]=MP[3*points[i]+2];
	}

      // mirroring of virtual points (across the boundary)
      data[0][ 5] = 2.0*MP[3*points[4]  ] - MP[3*points[2*valence+7]  ];
      data[1][ 5] = 2.0*MP[3*points[4]+1] - MP[3*points[2*valence+7]+1];
      data[2][ 5] = 2.0*MP[3*points[4]+2] - MP[3*points[2*valence+7]+2];
      
      data[0][ 6] = 2.0*MP[3*points[1]  ] - MP[3*points[0]  ];
      data[1][ 6] = 2.0*MP[3*points[1]+1] - MP[3*points[0]+1];
      data[2][ 6] = 2.0*MP[3*points[1]+2] - MP[3*points[0]+2];
      
      data[0][ 7] = 2.0*MP[3*points[2]  ] - MP[3*points[3]  ];
      data[1][ 7] = 2.0*MP[3*points[2]+1] - MP[3*points[3]+1];
      data[2][ 7] = 2.0*MP[3*points[2]+2] - MP[3*points[3]+2];
      
#ifdef SHARP_BOUNDARY_CORNER
      // sharp corner boundary
      data[0][ 8] = 4.0*MP[3*points[2]  ] +     MP[3*points[0]  ] -
	            2.0*MP[3*points[1]  ] - 2.0*MP[3*points[3]  ]; 
      data[1][ 8] = 4.0*MP[3*points[2]+1] +     MP[3*points[0]+1] -
                    2.0*MP[3*points[1]+1] - 2.0*MP[3*points[3]+1]; 
      data[2][ 8] = 4.0*MP[3*points[2]+2] +     MP[3*points[0]+2] -
      	            2.0*MP[3*points[1]+2] - 2.0*MP[3*points[3]+2]; 
#else
      // round corner boundary
      data[0][ 8] =-12.0*MP[3*points[2]  ] +     MP[3*points[0]  ] +
	             6.0*MP[3*points[1]  ] + 6.0*MP[3*points[3]  ]; 
      data[1][ 8] =-12.0*MP[3*points[2]+1] +     MP[3*points[0]+1] +
                     6.0*MP[3*points[1]+1] + 6.0*MP[3*points[3]+1]; 
      data[2][ 8] =-12.0*MP[3*points[2]+2] +     MP[3*points[0]+2] +
                     6.0*MP[3*points[1]+2] + 6.0*MP[3*points[3]+2]; 
#endif
            
      data[0][ 9] = 2.0*MP[3*points[2]  ] - MP[3*points[1]  ];
      data[1][ 9] = 2.0*MP[3*points[2]+1] - MP[3*points[1]+1];
      data[2][ 9] = 2.0*MP[3*points[2]+2] - MP[3*points[1]+2];
      
      data[0][10] = 2.0*MP[3*points[3]  ] - MP[3*points[0]  ];
      data[1][10] = 2.0*MP[3*points[3]+1] - MP[3*points[0]+1];
      data[2][10] = 2.0*MP[3*points[3]+2] - MP[3*points[0]+2];
      
      data[0][11] = 2.0*MP[3*points[12]  ] - MP[3*points[13]  ];
      data[1][11] = 2.0*MP[3*points[12]+1] - MP[3*points[13]+1];
      data[2][11] = 2.0*MP[3*points[12]+2] - MP[3*points[13]+2];
      
      for(i=12;i<2*valence+8;i++)
	{
	  assert(points[i]!=-1);
	  data[0][i]=MP[3*points[i]];
	  data[1][i]=MP[3*points[i]+1];
	  data[2][i]=MP[3*points[i]+2];
	}
    }
  else
    {
      cerr << "Error: Boundary_type="<<boundary_type<<" not yet implemented." << endl;
      exit(1);
    }

  //if(valence==4) EP=new BicubicPatch(data);
  //else
  EP=new FirstOrderDataDependentPatch(data,valence); // ********************

  //#define EXTRACT_PATCH
#ifdef EXTRACT_PATCH
  // extracting subdivisions of a patch for future analysis

  if(valence==4)
    {
      FILE *f;
      char fname[256];
      int i,j,k;
      double  P[200];
      double SP[200];

      for(i=0;i<2*valence+8;i++)
	{
	  P[3*i]  =MP[3*points[i]]  -MP[3*points[0]];
	  P[3*i+1]=MP[3*points[i]+1]-MP[3*points[0]+1];
	  P[3*i+2]=MP[3*points[i]+2]-MP[3*points[0]+2];
	}

      for(k=0;k<=10;k++)
	{

	  cout << "Saving patch "<<k<<endl;
	  sprintf(fname,"extra_patch_val_%d_sub_%d.pts",valence,k);
	  f=fopen(fname,"w");
	  assert(f);
	  fprintf(f,"# 0 = constrained, 1 = not constrained, 2 = boundary\n");
	  fprintf(f,"# format: index constraint valence x y z\n");
	  fprintf(f,"numpoints = %d\n",2*valence+8);
	  
	  for(i=0;i<2*valence+8;i++)
	    {
	      fprintf(f,"%d %d %d %1.18e %1.18e %1.18e\n",i,1,0,
		      P[3*i]  -P[0],
		      P[3*i+1]-P[1],
		      P[3*i+2]-P[2]);
	    }
	  fclose(f);
      
	  sprintf(fname,"extra_patch_val_%d_sub_%d.pat",valence,k);
	  f=fopen(fname,"w");
	  
	  fprintf(f,"%d ",valence);
	  for(i=0;i<2*valence+8;i++)
	    {
	      fprintf(f,"%d ",i);
	    }
	  fprintf(f,"\n");
	  
	  fclose(f);


	  for(i=0;i<3*(2*valence+8);i++)
	    {
	      SP[i]=0;
	    }

	  for(i=0;i<2*valence+8;i++)
	    {
	      for(j=0;j<2*valence+8;j++)
		{
		  //cout << Omega4[valence][(2*valence+8)*i+j] << " ";
		  SP[3*i]  +=Omega4[valence][(2*valence+8)*i+j]*P[3*j];
		  SP[3*i+1]+=Omega4[valence][(2*valence+8)*i+j]*P[3*j+1];
		  SP[3*i+2]+=Omega4[valence][(2*valence+8)*i+j]*P[3*j+2];
		}
	      //cout << endl;
	    }
	      
	  for(i=0;i<2*valence+8;i++)
	    {
	      P[3*i]  =SP[3*i]  -SP[3*0];
	      P[3*i+1]=SP[3*i+1]-SP[3*0+1];
	      P[3*i+2]=SP[3*i+2]-SP[3*0+2];
	    }

	}

      cout <<"Done!"<<endl;
      exit(1);
    }
#endif
  return EP;
}
//////////////////////////////////////////////////////////////////////////////
// create a patch with the coordinates given in MP, valence given by this object
void MeshPatch::update_Patch(double *MP)
{
  double *data[3];
  double tmp1[2*MAX_VALENCE+8];
  double tmp2[2*MAX_VALENCE+8];
  double tmp3[2*MAX_VALENCE+8];
  int i;
      
  data[0]=tmp1;
  data[1]=tmp2;
  data[2]=tmp3;
  
  cerr << "Error: Boundary -1." << endl;
  exit(1);
  for(i=0;i<2*valence+8;i++)
    {
      data[0][i]=MP[3*points[i]];
      data[1][i]=MP[3*points[i]+1];
      data[2][i]=MP[3*points[i]+2];
    }
  patch->update_parameters(data);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// create patch with data in MP, compute energy, delete it again
double MeshPatch::KmEnergy(double *MP)
{
  FirstOrderDataDependentPatch *P;
  double tmp;

  P=create_Patch(MP);
  tmp=P->KmEnergy();
  delete P;

  return(tmp);
}

//////////////////////////////////////////////////////////////////////////////
// compute the partial derivative of the energy of MP with respect to point (i,coord)
void MeshPatch::grad_KmEnergy(double *MP, double *grad)
{
  double G[3*(2*MAX_VALENCE+8)];
  FirstOrderDataDependentPatch *P;
  int j,k;

  P=create_Patch(MP);
  P->grad_KmEnergy(G);
  delete P;

  for(j=0;j<2*valence+8;j++)
    {
      if(points[j]!=-1)
	{
	  if(M->MPoints[points[j]]->constraint==1) // only add if not constrained
	    { // (if constrained gradient is zero)
	      for(k=0;k<3;k++)
		{
		  grad[3*points[j]+k]+=G[3*j+k];
		}
	    }
	}
    }
}

//////////////////////////////////////////////////////////////////////////////

void MeshPatch::Hess_KmEnergy(double *MP, DynamicCompCol_Mat_double *H)
{
  FirstOrderDataDependentPatch *P;
  double K[9*(2*MAX_VALENCE+8)*(2*MAX_VALENCE+8)];
  int i,j,k,l;

  P=create_Patch(MP);
  P->Hess_KmEnergy(K);
  delete P;

#ifdef DEBUG
  //cout << "valence " << valence << endl;

  // ************
  for(i=0;i<3*(2*valence+8);i++)
    {
      for(j=0;j<3*(2*valence+8);j++)
	{
	  if(fabs(K[i*3*(2*valence+8)+j]-K[j*3*(2*valence+8)+i])>1e-6)
	    cout <<"K "<< i << ", " << j << ", " << K[i*3*(2*valence+8)+j]
		 <<"!="<< K[j*3*(2*valence+8)+i] << endl;
	}
    }
  // ************
#endif

  for(i=0;i<2*valence+8;i++)
    {
      if(points[i]!=-1)
	{
	  if(M->MPoints[points[i]]->constraint==1) // only add if not constrained
	    {
	      for(j=0;j<2*valence+8;j++)
		{
		  if(points[j]!=-1)
		    {
		      if(M->MPoints[points[j]]->constraint==1) // only add if not constrained
			{
			  for(k=0;k<3;k++)
			    {
			      for(l=0;l<3;l++)
				{
				  //cout << i << ", " << j << ", " << k << ", " << l << endl; fflush(0);
				  H->add(3*points[i]+k,
					 3*points[j]+l, 
					 K[3*(2*valence+8)*(3*i+k) + 3*j+l]);
				  //cout << "+" << endl; fflush(0);
				}
			    }
			}
		    }
		}
	    }
	  else // if constrained, set diagonal to 1
	    {
	      // Idea: matrix should look like
	      //       xyz
	      //    ...000...
	      //    ...000...
	      //  x 000100000
	      //  y 000010000
	      //  z 000001000
	      //    ...000...  
	      
	      //cout << i << endl;
	      
	      H->set(3*points[i]+0,
		     3*points[i]+0, 
		     1.0);
	      H->set(3*points[i]+1,
		     3*points[i]+1, 
		     1.0);
	      H->set(3*points[i]+2,
		     3*points[i]+2, 
		     1.0);
	    }
	}
    }
}
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// create patch with data in MP, compute energy, delete it again
double MeshPatch::KpEnergy(double *MP)
{
  FirstOrderDataDependentPatch *P;
  double tmp;

  P=create_Patch(MP);
  tmp=P->KpEnergy();
  delete P;

  return(tmp);
}

//////////////////////////////////////////////////////////////////////////////
// compute the partial derivative of the energy of MP with respect to point (i,coord)
void MeshPatch::grad_KpEnergy(double *MP, double *grad)
{
  double G[3*(2*MAX_VALENCE+8)];
  FirstOrderDataDependentPatch *P;
  int j,k;

  P=create_Patch(MP);
#ifdef EXACT_DERIVATIVES
  P->grad_KpEnergy(G);
#else
  cerr << "Central Diff Grad" << endl;
  P->grad_KpEnergy_central_diff(G);
#endif
  
  delete P;

  // project gradient back from virtual to real points (if neccessary)
  if(boundary_type==1)
    {
      G[3*4+0]        +=2.0*G[3*5+0];
      G[3*4+1]        +=2.0*G[3*5+1];
      G[3*4+2]        +=2.0*G[3*5+2];
      G[3*(2*valence+7)+0]-=G[3*5+0];
      G[3*(2*valence+7)+1]-=G[3*5+1];
      G[3*(2*valence+7)+2]-=G[3*5+2];

      G[3*1+0]        +=2.0*G[3*6+0];
      G[3*1+1]        +=2.0*G[3*6+1];
      G[3*1+2]        +=2.0*G[3*6+2];
      G[3*0+0]            -=G[3*6+0];
      G[3*0+1]            -=G[3*6+1];
      G[3*0+2]            -=G[3*6+2];

      G[3*2+0]        +=2.0*G[3*7+0];
      G[3*2+1]        +=2.0*G[3*7+1];
      G[3*2+2]        +=2.0*G[3*7+2];
      G[3*3+0]            -=G[3*7+0];
      G[3*3+1]            -=G[3*7+1];
      G[3*3+2]            -=G[3*7+2];

      G[3*9+0]        +=2.0*G[3*8+0];
      G[3*9+1]        +=2.0*G[3*8+1];
      G[3*9+2]        +=2.0*G[3*8+2];
      G[3*10+0]           -=G[3*8+0];
      G[3*10+1]           -=G[3*8+1];
      G[3*10+2]           -=G[3*8+2];
    }
  if(boundary_type==2)
    {
      G[3*4+0]        +=2.0*G[3*5+0];
      G[3*4+1]        +=2.0*G[3*5+1];
      G[3*4+2]        +=2.0*G[3*5+2];
      G[3*(2*valence+7)+0]-=G[3*5+0];
      G[3*(2*valence+7)+1]-=G[3*5+1];
      G[3*(2*valence+7)+2]-=G[3*5+2];

      G[3*1+0]        +=2.0*G[3*6+0];
      G[3*1+1]        +=2.0*G[3*6+1];
      G[3*1+2]        +=2.0*G[3*6+2];
      G[3*0+0]            -=G[3*6+0];
      G[3*0+1]            -=G[3*6+1];
      G[3*0+2]            -=G[3*6+2];

      G[3*2+0]        +=2.0*G[3*7+0];
      G[3*2+1]        +=2.0*G[3*7+1];
      G[3*2+2]        +=2.0*G[3*7+2];
      G[3*3+0]            -=G[3*7+0];
      G[3*3+1]            -=G[3*7+1];
      G[3*3+2]            -=G[3*7+2];

#ifdef SHARP_BOUNDARY_CORNER

      G[3*0+0]        +=1.0*G[3*8+0]; // **********************
      G[3*0+1]        +=1.0*G[3*8+1];
      G[3*0+2]        +=1.0*G[3*8+2];
      G[3*1+0]        -=2.0*G[3*8+0];
      G[3*1+1]        -=2.0*G[3*8+1];
      G[3*1+2]        -=2.0*G[3*8+2];
      G[3*2+0]        +=4.0*G[3*8+0];
      G[3*2+1]        +=4.0*G[3*8+1];
      G[3*2+2]        +=4.0*G[3*8+2];
      G[3*3+0]        -=2.0*G[3*8+0];
      G[3*3+1]        -=2.0*G[3*8+1];
      G[3*3+2]        -=2.0*G[3*8+2];
      
#else

      G[3*0+0]        += 1.0*G[3*8+0]; // **********************
      G[3*0+1]        += 1.0*G[3*8+1];
      G[3*0+2]        += 1.0*G[3*8+2];
      G[3*1+0]        += 6.0*G[3*8+0];
      G[3*1+1]        += 6.0*G[3*8+1];
      G[3*1+2]        += 6.0*G[3*8+2];
      G[3*2+0]        -=12.0*G[3*8+0];
      G[3*2+1]        -=12.0*G[3*8+1];
      G[3*2+2]        -=12.0*G[3*8+2];
      G[3*3+0]        += 6.0*G[3*8+0];
      G[3*3+1]        += 6.0*G[3*8+1];
      G[3*3+2]        += 6.0*G[3*8+2];

#endif

      G[3*2+0]        +=2.0*G[3*9+0];
      G[3*2+1]        +=2.0*G[3*9+1];
      G[3*2+2]        +=2.0*G[3*9+2];
      G[3*1+0]            -=G[3*9+0];
      G[3*1+1]            -=G[3*9+1];
      G[3*1+2]            -=G[3*9+2];

      G[3*3+0]        +=2.0*G[3*10+0];
      G[3*3+1]        +=2.0*G[3*10+1];
      G[3*3+2]        +=2.0*G[3*10+2];
      G[3*0+0]            -=G[3*10+0];
      G[3*0+1]            -=G[3*10+1];
      G[3*0+2]            -=G[3*10+2];

      G[3*12+0]       +=2.0*G[3*11+0];
      G[3*12+1]       +=2.0*G[3*11+1];
      G[3*12+2]       +=2.0*G[3*11+2];
      G[3*13+0]           -=G[3*11+0];
      G[3*13+1]           -=G[3*11+1];
      G[3*13+2]           -=G[3*11+2];
    }

  // transfer local gradient vector into global vector
  for(j=0;j<2*valence+8;j++)
    {
      if(points[j]!=-1) // only if point is not virtual
	{
	  if(M->MPoints[points[j]]->constraint==1) // only add if not constrained
	    { // (if constrained gradient is zero)
	      for(k=0;k<3;k++)
		{
		  grad[3*points[j]+k]+=G[3*j+k];
		}
	    }
	}
    }
}

//////////////////////////////////////////////////////////////////////////////

void MeshPatch::Hess_KpEnergy(double *MP, DynamicCompCol_Mat_double *H)
{
  FirstOrderDataDependentPatch *P;
  double K[9*(2*MAX_VALENCE+8)*(2*MAX_VALENCE+8)];
  int i,j,k,l;

  if(all_points_constrained)
    {
      for(i=0;i<2*valence+8;i++)
	{
	  if(points[i]!=-1)
	    {
	      H->set(3*points[i]+0,
		     3*points[i]+0, 
		     1.0);
	      H->set(3*points[i]+1,
		     3*points[i]+1, 
		     1.0);
	      H->set(3*points[i]+2,
		     3*points[i]+2, 
		     1.0);
	    }
	}
    }
  else
    {
      P=create_Patch(MP);
#ifdef EXACT_DERIVATIVES
      P->Hess_KpEnergy(K);
#else
      cerr << "Central Diff Hess" << endl;
      P->Hess_KpEnergy_central_diff(K);
#endif
      delete P;

#ifdef DEBUG
      //cout << "valence " << valence << endl;

      // ************
      for(i=0;i<3*(2*valence+8);i++)
	{
	  for(j=0;j<3*(2*valence+8);j++)
	    {
	      if(fabs(K[i*3*(2*valence+8)+j]-K[j*3*(2*valence+8)+i])>1e-6)
		cout <<"K "<< i << ", " << j << ", " << K[i*3*(2*valence+8)+j]
		     <<"!="<< K[j*3*(2*valence+8)+i] << endl;
	    }
	}
      // ************
#endif

      // project Hessian back from virtual to real points (if neccessary)
      if(boundary_type==1)
	{
	  for(i=0;i<2*valence+8;i++)
	    {
	      for(k=0;k<3;k++)
		{	  
		  for(l=0;l<3;l++)
		    {	  
		      K[(3*i+k)*3*(2*valence+8)+(3* 4+l)]           +=2.0*K[(3*i+k)*3*(2*valence+8)+(3*5+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3*(2*valence+7)+l)]-=    K[(3*i+k)*3*(2*valence+8)+(3*5+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 1+l)]           +=2.0*K[(3*i+k)*3*(2*valence+8)+(3*6+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 0+l)]           -=    K[(3*i+k)*3*(2*valence+8)+(3*6+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 2+l)]           +=2.0*K[(3*i+k)*3*(2*valence+8)+(3*7+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 3+l)]           -=    K[(3*i+k)*3*(2*valence+8)+(3*7+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 9+l)]           +=2.0*K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3*10+l)]           -=    K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
		    }
		}
	    }
	  for(i=0;i<2*valence+8;i++)
	    {
	      for(k=0;k<3;k++)
		{	  
		  for(l=0;l<3;l++)
		    {	  
		      K[(3*i+k)+3*(2*valence+8)*(3* 4+l)]           +=2.0*K[(3*i+k)+3*(2*valence+8)*(3*5+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3*(2*valence+7)+l)]-=    K[(3*i+k)+3*(2*valence+8)*(3*5+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 1+l)]           +=2.0*K[(3*i+k)+3*(2*valence+8)*(3*6+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 0+l)]           -=    K[(3*i+k)+3*(2*valence+8)*(3*6+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 2+l)]           +=2.0*K[(3*i+k)+3*(2*valence+8)*(3*7+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 3+l)]           -=    K[(3*i+k)+3*(2*valence+8)*(3*7+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 9+l)]           +=2.0*K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3*10+l)]           -=    K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
		    }
		}
	    }
	}
      if(boundary_type==2)
	{
	  for(i=0;i<2*valence+8;i++)
	    {
	      for(k=0;k<3;k++)
		{	  
		  for(l=0;l<3;l++)
		    {	  
		      K[(3*i+k)*3*(2*valence+8)+(3* 4+l)]           += 2.0*K[(3*i+k)*3*(2*valence+8)+(3*5+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3*(2*valence+7)+l)]-=     K[(3*i+k)*3*(2*valence+8)+(3*5+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 1+l)]           += 2.0*K[(3*i+k)*3*(2*valence+8)+(3*6+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 0+l)]           -=     K[(3*i+k)*3*(2*valence+8)+(3*6+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 2+l)]           += 2.0*K[(3*i+k)*3*(2*valence+8)+(3*7+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 3+l)]           -=     K[(3*i+k)*3*(2*valence+8)+(3*7+l)];
#ifdef SHARP_BOUNDARY_CORNER
		      K[(3*i+k)*3*(2*valence+8)+(3* 0+l)]           +=     K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 1+l)]           -= 2.0*K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 2+l)]           += 4.0*K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 3+l)]           -= 2.0*K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
#else
		      K[(3*i+k)*3*(2*valence+8)+(3* 0+l)]           +=     K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 1+l)]           += 6.0*K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 2+l)]           -=12.0*K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 3+l)]           += 6.0*K[(3*i+k)*3*(2*valence+8)+(3*8+l)];
#endif
		      K[(3*i+k)*3*(2*valence+8)+(3* 2+l)]           += 2.0*K[(3*i+k)*3*(2*valence+8)+(3*9+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 1+l)]           -=     K[(3*i+k)*3*(2*valence+8)+(3*9+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 3+l)]           += 2.0*K[(3*i+k)*3*(2*valence+8)+(3*10+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3* 0+l)]           -=     K[(3*i+k)*3*(2*valence+8)+(3*10+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3*12+l)]           += 2.0*K[(3*i+k)*3*(2*valence+8)+(3*11+l)];
		      K[(3*i+k)*3*(2*valence+8)+(3*13+l)]           -=     K[(3*i+k)*3*(2*valence+8)+(3*11+l)];
		    }
		}
	    }
	  for(i=0;i<2*valence+8;i++)
	    {
	      for(k=0;k<3;k++)
		{	  
		  for(l=0;l<3;l++)
		    {	  
		      K[(3*i+k)+3*(2*valence+8)*(3* 4+l)]           += 2.0*K[(3*i+k)+3*(2*valence+8)*(3*5+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3*(2*valence+7)+l)]-=     K[(3*i+k)+3*(2*valence+8)*(3*5+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 1+l)]           += 2.0*K[(3*i+k)+3*(2*valence+8)*(3*6+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 0+l)]           -=     K[(3*i+k)+3*(2*valence+8)*(3*6+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 2+l)]           += 2.0*K[(3*i+k)+3*(2*valence+8)*(3*7+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 3+l)]           -=     K[(3*i+k)+3*(2*valence+8)*(3*7+l)];
#ifdef SHARP_BOUNDARY_CORNER
		      K[(3*i+k)+3*(2*valence+8)*(3* 0+l)]           +=     K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 1+l)]           -= 2.0*K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 2+l)]           += 4.0*K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 3+l)]           -= 2.0*K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
#else
		      K[(3*i+k)+3*(2*valence+8)*(3* 0+l)]           +=     K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 1+l)]           += 6.0*K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 2+l)]           -=12.0*K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 3+l)]           += 6.0*K[(3*i+k)+3*(2*valence+8)*(3*8+l)];
#endif
		      K[(3*i+k)+3*(2*valence+8)*(3* 2+l)]           += 2.0*K[(3*i+k)+3*(2*valence+8)*(3*9+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 1+l)]           -=     K[(3*i+k)+3*(2*valence+8)*(3*9+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 3+l)]           += 2.0*K[(3*i+k)+3*(2*valence+8)*(3*10+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3* 0+l)]           -=     K[(3*i+k)+3*(2*valence+8)*(3*10+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3*12+l)]           += 2.0*K[(3*i+k)+3*(2*valence+8)*(3*11+l)];
		      K[(3*i+k)+3*(2*valence+8)*(3*13+l)]           -=     K[(3*i+k)+3*(2*valence+8)*(3*11+l)];
		    }
		}
	    }
	}

      for(i=0;i<2*valence+8;i++)
	{
	  if(points[i]!=-1)
	    {
	      if(M->MPoints[points[i]]->constraint==1) // only add if not constrained
		{
		  for(j=0;j<2*valence+8;j++)
		    {
		      if(points[j]!=-1)
			{
			  if(M->MPoints[points[j]]->constraint==1) // only add if not constrained
			    {
			      for(k=0;k<3;k++)
				{
				  for(l=0;l<3;l++)
				    {
				      //cout << i << ", " << j << ", " << k << ", " << l << endl; fflush(0);
				      H->add(3*points[i]+k,
					     3*points[j]+l, 
					     K[3*(2*valence+8)*(3*i+k) + 3*j+l]);
				      //cout << "+" << endl; fflush(0);
				    }
				}
			    }
			}
		    }
		}
	      else // if constrained, set diagonal to 1
		{
		  // Idea: matrix should look like
		  //       xyz
		  //    ...000...
		  //    ...000...
		  //  x 000100000
		  //  y 000010000
		  //  z 000001000
		  //    ...000...  
		  
		  //cout << i << endl;
		  
		  H->set(3*points[i]+0,
			 3*points[i]+0, 
			 1.0);
		  H->set(3*points[i]+1,
			 3*points[i]+1, 
			 1.0);
		  H->set(3*points[i]+2,
			 3*points[i]+2, 
			 1.0);
		}
	    }
	}
    }
}

//////////////////////////////////////////////////////////////////////////////

void MeshPatch::Hess_KpEnergy_mult_vector(double *X, double *A, double *result)
{
  FirstOrderDataDependentPatch *P;
  double K[9*(2*MAX_VALENCE+8)*(2*MAX_VALENCE+8)];
  int i,j,k,l;

  P=create_Patch(X);
  P->Hess_KpEnergy(K);
  delete P;

  cerr << "Error: Boundary -1." << endl;
  exit(1);
  for(i=0;i<2*valence+8;i++)
    {
      if(M->MPoints[points[i]]->constraint==1) // only add if not constrained
	{
	  for(j=0;j<2*valence+8;j++)
	    {
	      if(M->MPoints[points[j]]->constraint==1) // only add if not constrained
		{
		  for(k=0;k<3;k++)
		    {
		      for(l=0;l<3;l++)
			{
			  result[3*points[i]+k]+=K[3*(2*valence+8)*(3*i+k) + 3*j+l]*A[3*points[j]+l];
			}
		    }
		}
	    }
	}
      else
	{
	  // Lagrangian multiplier - just copy values
	  result[3*points[i]+0] = A[3*points[i]+0];
	  result[3*points[i]+1] = A[3*points[i]+1];
	  result[3*points[i]+2] = A[3*points[i]+2];
	}
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// ********* delete all MeshPoint arrays **********
Mesh::~Mesh()
{
  int i;

  MeshPatch *MP;
  MeshPatch *tmp;

  for(i=0;i<numpoints;i++) 
    {
      if(MPoints[i]) delete  MPoints[i];
      MPoints[i]=NULL;
    }
  delete[] MPoints;
  MPoints=NULL;

  if(P) delete[] P;
  P=NULL;

  cout << "~Mesh: Points deleted"<<endl;
  fflush(0);

  MP=firstpatch;
  while(MP)
    {
      tmp=MP;
      MP =MP->next;
      delete tmp;
    }
  cout << "~Mesh: Patches deleted"<<endl;
  fflush(0);
}

//////////////////////////////////////////////////////////////////////////////
// ok
Mesh::Mesh(char *fname)
{
  export_iv_sphere_size    = 1.0;
  export_iv_control_grid   = 1;
  export_iv_M              = 3;
  export_iv_N              = 3;
  export_iv_coloring_scheme= 0;
  export_iv_coloring_range = 10.0;
  export_iv_reflection_map = 0;

  P   = NULL;
  load(fname);
}

//////////////////////////////////////////////////////////////////////////////
// load mesh in Patrick and Ilja's format
int Mesh::load(char *fname)
{
  FILE *f;
  char tmp[1024];
  int i,ok;
  MeshPatch *MP; 

  sprintf(tmp,"%s.pts",fname);
  cout << "Opening <"<<tmp<<">"<<endl;
  f=fopen(tmp,"r");
  assert(f);
  fscanf(f,"#%[^#]s\n",tmp);
  cout << "Reading <"<<tmp<<">"<<endl;
  fscanf(f,"#%[^\n]s\n",tmp);
  cout << "Reading <"<<tmp<<">"<<endl;
  fscanf(f,"\nnumpoints = %d\n",&numpoints);
  cout << "numpoints = "<<numpoints<<endl;

  MPoints=new MeshPoint*[numpoints];
  P      =new double[3*numpoints];

  for(i=0;i<numpoints;i++)
    {
      MPoints[i]=new MeshPoint(f,P,i);
    }
  fclose(f);

  sprintf(tmp,"%s.pat",fname);
  cout << "Opening " << tmp << endl;
  f=fopen(tmp,"r");
  assert(f);

  firstpatch=new MeshPatch(this,f);//,P);
  MP=firstpatch;

  if(firstpatch->valence) ok=1;
  else
    {
      ok=0;
      delete firstpatch;
    }
      
  while(ok)
    {
      MP->next=new MeshPatch(this,f);//,P);
      numpatches++;

      if(MP->next->valence==0)
	{
	  delete(MP->next);
	  MP->next=NULL;
	  ok=0;
	}
      else 
	{
	  MP=MP->next;	  
	  ok=1;
	}
    }
  
  fclose(f);
  set_connectivity_information();
  return(1);
}

//////////////////////////////////////////////////////////////////////////////
// save mesh in Patrick and Ilja's format
int Mesh::save(char *fname)
{
  FILE *f;
  char tmp[1024];
  int i;
  MeshPatch *MP; 

  cout << "Saving mesh <"<<fname<<">"<<endl;
  sprintf(tmp,"%s.pts",fname);
  f=fopen(tmp,"w");
  assert(f);
  fprintf(f,"# 0 = constrained, 1 = not constrained, 2 = boundary\n");
  fprintf(f,"# format: index constraint valence x y z\n");
  fprintf(f,"numpoints = %d\n",numpoints);

  for(i=0;i<numpoints;i++)
    {
      MPoints[i]->save(f,P);
    }
  fclose(f);

  sprintf(tmp,"%s.pat",fname);
  f=fopen(tmp,"w");
  assert(f);
  MP=firstpatch;
  while(MP)
    {
      MP->save(f);
      MP=MP->next;
    }
  fclose(f);
  return(1);
}

//////////////////////////////////////////////////////////////////////////////
// write mesh to file in IV format
int Mesh::export_IV(char *fname, double *P)
{
  MeshPatch *MP;
  int i;

  cout << "export_IV('"<<fname<<"')"<<endl;

  MP=firstpatch;
  while(MP)
    {
      MP->export_IV(fname,P);
      MP=MP->next;
    }

  for(i=0;i<numpoints;i++) MPoints[i]->export_IV(fname,P);

  return(1);
}
//////////////////////////////////////////////////////////////////////////////
// write mesh to file in IV format
int Mesh::export_IV(char *fname, double *P, double *grad)
{
  FILE *f;
  int i;
  int grad_is_zero;
  
  export_IV(fname,P);

  grad_is_zero=1;
  for(i=0;i<3*numpoints;i++) if(grad[i]!=0.0) grad_is_zero=0;
  
  if(!grad_is_zero)
    {
      f=fopen(fname,"a+");
      assert(f);
  
      fprintf(f,"   Separator {\n");
      fprintf(f,"           Material {\n");
      fprintf(f,"                ambientColor   0.5 0.5 0.5 diffuseColor 0.5 0.5 0.5   # Red\n");
      fprintf(f,"           }\n");
      fprintf(f,"        Coordinate3 {\n");
      fprintf(f,"            point [\n");
      
      for(i=0;i<numpoints;i++) 
	{
	  fprintf(f,"%e %e %e, ",
		  P[3*i+0], 
		  P[3*i+1], 
		  P[3*i+2]);
	  
	  fprintf(f,"%e %e %e",
		  P[3*i+0]+grad[3*i+0], 
		  P[3*i+1]+grad[3*i+1], 
		  P[3*i+2]+grad[3*i+2]);
	  
	  if(i<numpoints-1) fprintf(f,", \n");
	  else fprintf(f,"\n");
	}
      
      fprintf(f,"                  ]\n");
      fprintf(f,"        }\n");
      fprintf(f,"        IndexedLineSet {\n");
      fprintf(f,"            coordIndex [ ");
      for(i=0;i<numpoints;i++) 
	{
	  if(i<numpoints-1) fprintf(f,"%d, %d, -1,\n",2*i,2*i+1);
	  else fprintf(f,"%d, %d, -1\n",2*i,2*i+1);
	}
      fprintf(f,"                       ]\n");
      fprintf(f,"        }");
      fprintf(f,"    }\n");
      fclose(f);
    }

  return(1);
}

//////////////////////////////////////////////////////////////////////////////
// sets connectivity information and all_points_constrained flag
void Mesh::set_connectivity_information(void)
{
  MeshPatch *MP;
  int i,j;

  // go over all points and tell them how many patches they control
  MP=firstpatch;
  while(MP)
    {
      for(i=0;i<2*MP->valence+8;i++)
	{
	  if(MP->points[i]!=-1) MPoints[MP->points[i]]->patches_in_ring++;	  	    
	}
      MP->all_points_constrained=1; // temporary guess
      MP=MP->next;
    }

  // create for each point a table with references to the participating patches
  for(i=0;i<numpoints;i++)
    {
      MPoints[i]->ring_table=new MeshPatch*[MPoints[i]->patches_in_ring];
      for(j=0;j<MPoints[i]->patches_in_ring;j++) MPoints[i]->ring_table[j]=NULL;
    }

  // for all patches
  MP=firstpatch;
  while(MP)
    {
      // for all the points in the patch
      for(i=0;i<2*MP->valence+8;i++)
	{
	  if(MP->points[i]!=-1)
	    {
	      j=0;
	      while(MPoints[MP->points[i]]->ring_table[j]) j++;
	      assert(j<MPoints[MP->points[i]]->patches_in_ring);
	      
	      MPoints[MP->points[i]]->ring_table[j]=MP;
	      
	      // constraint type 0 is fixed, all other types are assumed to be free
	      if(MPoints[MP->points[i]]->constraint!=0) MP->all_points_constrained=0;
	    }
	}
      MP=MP->next;
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// to do: check for all_points_constrained *****
double Mesh::KmEnergy(double *A)
{
  double tmp;
  MeshPatch *MP;

  tmp=0.0;
  MP=firstpatch;
  while(MP)
    {
      tmp+=MP->KmEnergy(A); // sum over all energies
      MP=MP->next;
    }
  return(tmp);
}

//////////////////////////////////////////////////////////////////////////////
// compute the gradient grad of the energy of mesh A
// constraints are already taken care of in MeshPatch
// to do: check for all_points_constrained *****

void Mesh::grad_KmEnergy(double *A, double *grad)
{
  int       i;
  MeshPatch *MP;

  for(i=0;i<3*numpoints;i++)
    {
      grad[i]=0.0;
    }

  MP=firstpatch;
  while(MP)
    {
      MP->grad_KmEnergy(A,grad); 
      MP=MP->next;
    }
}
//////////////////////////////////////////////////////////////////////////////
// compute the hessian Hess of the energy of mesh A
// constraints are already taken care of in MeshPatch
// to do: check for all_points_constrained *****

DynamicCompCol_Mat_double* Mesh::Hess_KmEnergy(double *A)
{
  MeshPatch *tmp;
  DynamicCompCol_Mat_double *H;

  H=new DynamicCompCol_Mat_double(3*numpoints);

  tmp=firstpatch;
  while(tmp)
    {
      tmp->Hess_KmEnergy(A,H);
      tmp=tmp->next;
    }
  return(H);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double Mesh::KpEnergy(double *A)
{
  double tmp;
  MeshPatch *MP;

  tmp=0.0;
  MP=firstpatch;
  while(MP)
    {
      if(!(MP->all_points_constrained)) tmp+=MP->KpEnergy(A); // sum over all energies that could change
      MP=MP->next;
    }
  return(tmp);
}

//////////////////////////////////////////////////////////////////////////////
// compute the gradient grad of the energy of mesh A
// constraints are already taken care of in MeshPatch

void Mesh::grad_KpEnergy(double *A, double *grad)
{
  int       i;
  MeshPatch *MP;

  for(i=0;i<3*numpoints;i++)
    {
      grad[i]=0.0;
      i++;
      grad[i]=0.0;
      i++;
      grad[i]=0.0;
    }

  MP=firstpatch;
  while(MP)
    {
      if(!(MP->all_points_constrained)) MP->grad_KpEnergy(A,grad); 
      MP=MP->next;
    }
}

//////////////////////////////////////////////////////////////////////////////
// compute the hessian Hess of the energy of mesh A
// constraints are already taken care of in MeshPatch

DynamicCompCol_Mat_double* Mesh::Hess_KpEnergy(double *A)
{
  MeshPatch *tmp;
  DynamicCompCol_Mat_double *H;

  H=new DynamicCompCol_Mat_double(3*numpoints);

  tmp=firstpatch;
  while(tmp)
    {
      tmp->Hess_KpEnergy(A,H); // even if all points constrained -> handling of Lagrangian multipliers internally
      tmp=tmp->next;
    }
  //H->clear_from_zeros();  *********************
  return(H);
}

//////////////////////////////////////////////////////////////////////////////
// Hess[Kp](X)*A=result
// only efficient if a single matrix vector multiplication is needed

void Mesh::Hess_KpEnergy_mult_vector(double *X, double *A, double *result)
{
  MeshPatch *tmp;
  int i;

  // clear vector
  for(i=0;i<3*numpoints;i++)
    {
      result[i]=0.0;
      i++;
      result[i]=0.0;
      i++;
      result[i]=0.0;
    }

  // multiply each patch with vector
  tmp=firstpatch;
  while(tmp)
    {
      // even if all points constrained -> handling 
      // of Lagrangian multipliers internally
      tmp->Hess_KpEnergy_mult_vector(X,A,result); 
      tmp=tmp->next;
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
