//////////////////////////////////////////////////////////////////////////////
//
// BicubicPatch.h, 07/10/01, Ilja Friedel
//
//
//////////////////////////////////////////////////////////////////////////////

#ifndef BICUBICPATCH_USED
#define BICUBICPATCH_USED

#include <assert.h>
#include <stdio.h>
#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;
#include "Patch.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

class BicubicPatch:public Patch
{
  static int    initialized;
  static double eX_tensor[256];
  static double iX_tensor[256];
  static double eX_patrick[256];
  static double iX_patrick[256];
  static double Lambda[256];

  //-1/384*p00-23/384*p01-23/384*p02-1/384*p03-5/384*p04-115/384*p05-115/384*p06-5/384*p07+5/384*p08+115/384*p09+115/384*p10+5/384*p11+1/384*p12+23/384*p13+23/384*p14+1/384*p15

  //-1/384*p00-5/384*p01+5/384*p02+1/384*p03-23/384*p04-115/384*p05+115/384*p06+23/384*p07-23/384*p08-115/384*p09+115/384*p10+23/384*p11-1/384*p12-5/384*p13+5/384*p14+1/384*p15

  void set_matrices(void);
  void load_matrix(char *fname, double *p, int n, int m);

//////////////////////////////////////////////////////////////////////////////

  double *p[3];  // control points
  double *A[3];  // horner coefficients of bicubic patch 

 public:

  double Kp_function(const double u, const double v);
  double Km_function(const double u, const double v);

//////////////////////////////////////////////////////////////////////////////

  double evaluate_S(const double u, const double v, const int s);

  double evaluate_D1S(const double u, const double v, const int s);
  double evaluate_D2S(const double u, const double v, const int s);

  double evaluate_D1D1S(const double u, const double v, const int s);
  double evaluate_D1D2S(const double u, const double v, const int s);
  double evaluate_D2D1S(const double u, const double v, const int s);
  double evaluate_D2D2S(const double u, const double v, const int s);

//////////////////////////////////////////////////////////////////////////////

  void add_sample_to_Kp(const double u, const double v, const double weight, double *K);
  void add_sample_to_Km(const double u, const double v, const double weight, double *K);

  void   Kp(Patch *S, double *K);
  void   Km(Patch *S, double *K);

//////////////////////////////////////////////////////////////////////////////

  void print(void);

//////////////////////////////////////////////////////////////////////////////

  BicubicPatch(double *data);
  BicubicPatch(double **data);
  virtual ~BicubicPatch();
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#endif


