//////////////////////////////////////////////////////////////////////////////
//
// FirstOrderDataDependentPatch.h, 11/02/01, Ilja Friedel
//
//////////////////////////////////////////////////////////////////////////////

#ifndef FIRSTORDERDATADEPENDENTPATCH_USED
#define FIRSTORDERDATADEPENDENTPATCH_USED

//////////////////////////////////////////////////////////////////////////////

#include <assert.h>
#include <stdio.h>
#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;
#include "Patch.h"
#include "BicubicPatch.h"
#include "EigenStructure.h"
#include "ExtraordinaryPatch.h"

//////////////////////////////////////////////////////////////////////////////

extern double *FODD_Kp[6][MAX_VALENCE+1];
extern double *FODD_Km[4][MAX_VALENCE+1];

//////////////////////////////////////////////////////////////////////////////

class FirstOrderDataDependentPatch:public ExtraordinaryPatch
{
  static int initialized;

  // xx and xxxx are very sensitive "global" variables 

  int xx;    // for membrane energy function
             // xx=0 -> XX  new E
             // xx=1 -> XY  new F
             // xx=2 -> YY  new G
             // xx=3 -> ZZ
  int xxxx;  // for bending energy function
             // xxxx=0 -> EE
             // xxxx=1 -> EF
             // xxxx=2 -> EG
             // xxxx=3 -> FF
             // xxxx=4 -> FG
             // xxxx=5 -> GG

  int  load_Kp(void);
  void save_Kp(void);

  int  load_Km(void);
  void save_Km(void);

	  //map2(subs,{u=1/2,v=1/2},d1S);
	  //115/384*p00-115/384*p01-115/384*p02+115/384*p03-5/384*p04-1/384*p05-23/384*p06-23/384*p07-1/384*p08-5/384*p09+5/384*p10+1/384*p11+23/384*p12+23/384*p13+1/384*p14+5/384*p15

	  //map2(subs,{u=1/2,v=1/2},d2S);
	  //115/384*p00+115/384*p01-115/384*p02-115/384*p03+23/384*p04+1/384*p05+5/384*p06-5/384*p07-1/384*p08-23/384*p09-23/384*p10-1/384*p11-5/384*p12+5/384*p13+1/384*p14+23/384*p15

  static const double VdS4[2][16];
  /*
 = { 
	      { 115./384., 
	       -115./384.,
	       -115./384.,
	        115./384.,
		 -5./384.,
		 -1./384.,
		-23./384.,
		-23./384.,
		 -1./384.,
		 -5./384.,
  		  5./384.,
		  1./384.,
		 23./384.,
		 23./384.,
		  1./384.,
		  5./384.},
	      {
		115./384.,
		115./384.,
	       -115./384.,
	       -115./384.,
		 23./384.,
		  1./384.,
		  5./384.,
		 -5./384.,
		 -1./384.,
		-23./384.,
		-23./384.,
		 -1./384.,
		 -5./384.,
		  5./384.,
 		  1./384.,
		 23./384.}};
  */
 public:

  double Kp_function(const double, const double);
  double Km_function(const double, const double);
  // *************************************************************************
  // *************************************************************************
  // *************************************************************************
  // *************************************************************************
  // recompute all Kp and Km matrices on disc
  // *************************************************************************
  // *************************************************************************
  // *************************************************************************

//////////////////////////////////////////////////////////////////////////////

  void   Kp(Patch *S, double *K);
  void   Km(Patch *S, double *K);

//////////////////////////////////////////////////////////////////////////////
  virtual double KpEnergy(void);
  virtual void   grad_KpEnergy(double *G);
  virtual void   Hess_KpEnergy(double *H);

  virtual void   grad_KpEnergy_central_diff(double *G);
  virtual void   Hess_KpEnergy_central_diff(double *H);

//////////////////////////////////////////////////////////////////////////////
  virtual double KmEnergy(void);
  virtual void   grad_KmEnergy(double *G);
  virtual void   Hess_KmEnergy(double *H);

  virtual void   grad_KmEnergy_central_diff(double *G);
  virtual void   Hess_KmEnergy_central_diff(double *H);

//////////////////////////////////////////////////////////////////////////////

  double query_Sx(void);
  double query_Sxy(void);
  double query_Sz(void);

  double query_E(void);
  double query_F(void);
  double query_G(void);

//////////////////////////////////////////////////////////////////////////////

  void save_membrane_integrands(int,int,int);
  void save_bending_integrands(int,int,int);

//////////////////////////////////////////////////////////////////////////////

  virtual ~FirstOrderDataDependentPatch();
  FirstOrderDataDependentPatch(double **data, int N);
};

//////////////////////////////////////////////////////////////////////////////

#endif
