//////////////////////////////////////////////////////////////////////////////
//
// FirstOrderDataDependentPatch.cpp, 05/19/02, Ilja Friedel
//
//////////////////////////////////////////////////////////////////////////////

#include "FirstOrderDataDependentPatch.h"

//////////////////////////////////////////////////////////////////////////////

double *FODD_Kp[6][MAX_VALENCE+1];
double *FODD_Km[4][MAX_VALENCE+1];
int FirstOrderDataDependentPatch::initialized=0;

//////////////////////////////////////////////////////////////////////////////
//
// The weights for the linear combination of points, needed to compute
// the first partial derivatives of the parameterized surface.
//
// map2(subs,{u=1/2,v=1/2},d1S);
//   115/384*p00-115/384*p01-115/384*p02+115/384*p03
//    -5/384*p04-1/384*p05-23/384*p06-23/384*p07
//    -1/384*p08-5/384*p09+5/384*p10+1/384*p11
//   +23/384*p12+23/384*p13+1/384*p14+5/384*p15
//
// map2(subs,{u=1/2,v=1/2},d2S);
//   115/384*p00+115/384*p01-115/384*p02-115/384*p03
//   +23/384*p04+1/384*p05+5/384*p06-5/384*p07
//    -1/384*p08-23/384*p09-23/384*p10
//    -1/384*p11-5/384*p12+5/384*p13+1/384*p14+23/384*p15
//
// ************************************************************
// Had to switch both vectors!!!!
// ************************************************************
//////////////////////////////////////////////////////////////////////////////

const double FirstOrderDataDependentPatch::VdS4[2][16] = { 
	      {
		115./384.,
		115./384.,
	       -115./384.,
	       -115./384.,
		 23./384.,
		  1./384.,
		  5./384.,
		 -5./384.,
		 -1./384.,
		-23./384.,
		-23./384.,
		 -1./384.,
		 -5./384.,
		  5./384.,
 		  1./384.,
		 23./384.}
	      ,
	      { 115./384., 
	       -115./384.,
	       -115./384.,
	        115./384.,
		 -5./384.,
		 -1./384.,
		-23./384.,
		-23./384.,
		 -1./384.,
		 -5./384.,
  		  5./384.,
		  1./384.,
		 23./384.,
		 23./384.,
		  1./384.,
		  5./384.},
               };

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

FirstOrderDataDependentPatch::~FirstOrderDataDependentPatch()
{
  if(control_points[0]) delete[] control_points[0];
  if(control_points[1]) delete[] control_points[1];
  if(control_points[2]) delete[] control_points[2];

  control_points[0]=NULL;
  control_points[1]=NULL;
  control_points[2]=NULL;

  if(ES) delete ES;
  ES=NULL;
}

//////////////////////////////////////////////////////////////////////////////

FirstOrderDataDependentPatch::FirstOrderDataDependentPatch(double **data, int N):ExtraordinaryPatch(data,N)
{
  int i;
  
  xxxx=0;
  if(!FirstOrderDataDependentPatch::initialized)
    {
      FirstOrderDataDependentPatch::initialized=1;

      for(i=0;i<MAX_VALENCE;i++)
	{
	  FODD_Kp[0][i]=NULL;
	  FODD_Km[0][i]=NULL;
	}
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double FirstOrderDataDependentPatch::query_Sx(void)
{
  cerr << "Error: Not implemented" << endl;
  exit(1);
  return(0.0);
}

//////////////////////////////////////////////////////////////////////////////

double FirstOrderDataDependentPatch::query_Sxy(void)
{
  cerr << "Error: Not implemented" << endl;
  exit(1);
  return(0.0);
}

//////////////////////////////////////////////////////////////////////////////

double FirstOrderDataDependentPatch::query_Sz(void)
{
  cerr << "Error: Not implemented" << endl;
  exit(1);
  return(0.0);
}

//////////////////////////////////////////////////////////////////////////////

double FirstOrderDataDependentPatch::query_E(void)
{
  cerr << "Error: Not implemented" << endl;
  exit(1);
  return(0.0);
}

//////////////////////////////////////////////////////////////////////////////

double FirstOrderDataDependentPatch::query_F(void)
{
  cerr << "Error: Not implemented" << endl;
  exit(1);
  return(0.0);
}

//////////////////////////////////////////////////////////////////////////////

double FirstOrderDataDependentPatch::query_G(void)
{
  cerr << "Error: Not implemented" << endl;
  exit(1);
  return(0.0);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

