//////////////////////////////////////////////////////////
// QuadTree.C, 2/14/2000, Ilja Friedel, cs 174
//
// class QuadTree
//////////////////////////////////////////////////////////

#include "QuadTree.h"

#define QUADNODE_BORDER
//#define QUADNODE_INTERPOLATED

//////////////////////////////////////////////////////////

int QuadNode::OGLID(void){return(1);}
int LinkNode::OGLID(void){return(2);}

//////////////////////////////////////////////////////////

void LinkNode::OpenGL_draw(void)
{
  Vector a;
  Vector b;

  if(this==NULL) 
    {
      printf("Warning: tried to call (LinkNode *)NULL->OpenGL_draw()!\n");
    }

  a=q->Quad->corner+(q->x+0.5*q->dx)*q->Quad->dx+(q->y+0.5*q->dy)*q->Quad->dy;
  b=p->Quad->corner+(p->x+0.5*p->dx)*p->Quad->dx+(p->y+0.5*p->dy)*p->Quad->dy;
  
  glLoadName((unsigned int)this);
  
  glBegin(GL_LINES);
  if(highlight) glColor3d(0.5,0.7,0.3);
  else glColor3d(1.0,0.0,1.0);
  glVertex3d(a.X(),a.Y(),a.Z());
  glVertex3d(b.X(),b.Y(),b.Z());
  glEnd();

  if(next) next->OpenGL_draw();
}

//////////////////////////////////////////////////////////

void LinkNode::print_info(void)
{
  printf("Fpq=%e\n",Fpq);
  output(80,100,"Fpq=%e\n",Fpq);
}

//////////////////////////////////////////////////////////

int LinkNode::respond_click(int O)
{
  if((int)this==O)
    {
      highlight=1;
      if(q->L) q->L->OpenGL_draw();
      if(next) next->respond_click(O);
      return(1);
    }
  else
    {
      highlight=0;
      if(next) return(next->respond_click(O));
    }
  return(0);
}

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

Vector QuadNode::get_point(double a, double b)
{
  return(Quad->corner+
	 (x+a*dx)*Quad->dx+
	 (y+b*dy)*Quad->dy);
}

//////////////////////////////////////////////////////////

void QuadTree::highlight_branch(QuadNode *QN)
{
  if(QN->ru) QN->ru->highlight_branch();
  if(QN->lu) QN->lu->highlight_branch();
  if(QN->ll) QN->ll->highlight_branch();
  if(QN->rl) QN->rl->highlight_branch();
  
  while(QN)
    {
      QN->highlight_branch();
      if(QN->L) QN->L->OpenGL_draw();
      printf("Bs.red[%d]=%e Bs.green[%d]=%e Bs.blue[%d]=%e\n",
	     QN->level,QN->Bs.Red(),
	     QN->level,QN->Bs.Green(),
	     QN->level,QN->Bs.Blue());
      QN=QN->parent;
    }
}

//////////////////////////////////////////////////////////

void QuadNode::highlight_branch(void)
{
  Color pigment_rgb;
  Vector w,u,v,z;
  double red,green,blue;
  
  if(!Quad)
    {
      printf("Error: Quad is NULL!\n");
      printf("[void QuadNode::highlight_branch(void)]\n");
      exit(1);
    }

  red  =Quad->pigment_rgb.Red();
  green=Quad->pigment_rgb.Green();
  blue =Quad->pigment_rgb.Blue();
  
  red  =red  +0.5;
  green=green+0.5;
  blue =blue +0.5;
  
  if(red  >1.0) red  -=1.0;
  if(green>1.0) green-=1.0;
  if(blue >1.0) blue -=1.0;
  
  w=Quad->corner+x*Quad->dx+y*Quad->dy;
  u=w+dx*Quad->dx;
  v=w+dx*Quad->dx+dy*Quad->dy;
  z=w+dy*Quad->dy;
  
  glDepthFunc(GL_ALWAYS);
  glBegin(GL_LINE_STRIP);
  glColor3d(red,green,blue);
  glVertex3d(w.X(),w.Y(),w.Z());
  glVertex3d(u.X(),u.Y(),u.Z());
  glVertex3d(v.X(),v.Y(),v.Z());
  glVertex3d(z.X(),z.Y(),z.Z());
  glVertex3d(w.X(),w.Y(),w.Z());
  glEnd();

  glDepthFunc(GL_LESS);
}

//////////////////////////////////////////////////////////

void QuadNode::OpenGL_draw(void)
{
  if(lu && ll && ru && rl)
    {
      lu->OpenGL_draw();
      ll->OpenGL_draw();
      ru->OpenGL_draw();
      rl->OpenGL_draw();
    }
  else
    {
      Color pigment_rgb,C;
      Vector w,u,v,z;
      double red0,  red1,  red2,  red3,
	     green0,green1,green2,green3,
	     blue0, blue1, blue2, blue3;
      
      if(!Quad)
	{
	  printf("Error: Quad is NULL!\n");
	  printf("[void QuadNode::OpenGL_draw(void)]\n");
	  exit(1);
	}

      C=get_Color('S');
      red0  =C.Red();
      green0=C.Green();
      blue0 =C.Blue();
      if(red0  >1.0) red0  =1.0;
      if(green0>1.0) green0=1.0;
      if(blue0 >1.0) blue0 =1.0;

      C=get_Color('E');
      red1  =C.Red();
      green1=C.Green();
      blue1 =C.Blue();
      if(red1  >1.0) red1  =1.0;
      if(green1>1.0) green1=1.0;
      if(blue1 >1.0) blue1 =1.0;

      C=get_Color('N');
      red2  =C.Red();
      green2=C.Green();
      blue2 =C.Blue();
      if(red2  >1.0) red2  =1.0;
      if(green2>1.0) green2=1.0;
      if(blue2 >1.0) blue2 =1.0;

      C=get_Color('W');
      red3  =C.Red();
      green3=C.Green();
      blue3 =C.Blue();
      if(red3  >1.0) red3  =1.0;
      if(green3>1.0) green3=1.0;
      if(blue3 >1.0) blue3 =1.0;


      w=Quad->corner+x*Quad->dx+y*Quad->dy;
      u=w+dx*Quad->dx;
      v=w+dx*Quad->dx+dy*Quad->dy;
      z=w+dy*Quad->dy;
      
      glLoadName((unsigned int)this);
	  
      glPolygonMode(GL_FRONT,GL_FILL);
      glPolygonMode(GL_BACK,GL_LINE);
      glEnable(GL_CULL_FACE);
      glCullFace(GL_BACK);

      glBegin(GL_POLYGON);
      glColor3d(red0,green0,blue0);
      glVertex3d(w.X(),w.Y(),w.Z());
      glColor3d(red1,green1,blue1);
      glVertex3d(u.X(),u.Y(),u.Z());
      glColor3d(red2,green2,blue2);
      glVertex3d(v.X(),v.Y(),v.Z());
      glColor3d(red3,green3,blue3);
      glVertex3d(z.X(),z.Y(),z.Z());
      glEnd();

      glDisable(GL_CULL_FACE);

      // for borders

#ifdef QUADNODE_BORDER
      red0  =red0  +0.5;
      green0=green0+0.5;
      blue0 =blue0 +0.5;

      if(red0  >1.0) red0  =red0  -1.0;
      if(green0>1.0) green0=green0-1.0;
      if(blue0 >1.0) blue0 =blue0 -1.0;

      glDepthFunc(GL_ALWAYS);
      glBegin(GL_LINE_STRIP);
      glColor3d(red0,green0,blue0);
      glVertex3d(w.X(),w.Y(),w.Z());
      glVertex3d(u.X(),u.Y(),u.Z());
      glVertex3d(v.X(),v.Y(),v.Z());
      glVertex3d(z.X(),z.Y(),z.Z());
      glVertex3d(w.X(),w.Y(),w.Z());
      glEnd();
#endif
      glDepthFunc(GL_LESS);
    }
}

//////////////////////////////////////////////////////////

void QuadNode::print_info(void)
{
  output(80,100,"Bg=%1.3e Bs=%1.3e area=%1.2e E=%1.2e rho=%1.2e\n",
	 Bg.Red(),Bs.Red(),area,E.Red(),rho.Red());
  printf("For this node and red light: Bg=%e  Bs=%e  area=%e  E=%e  rho=%e\n",
	 Bg.Red(),Bs.Red(),area,E.Red(),rho.Red());
}

//////////////////////////////////////////////////////////

int QuadNode::respond_click(int O)
{
  if((int)this==O)
    {
      highlight=1;
      
      if(L) L->OpenGL_draw();
  
      if(ll) ll->respond_click(O);
      if(lu) lu->respond_click(O);
      if(ru) ru->respond_click(O);
      if(rl) rl->respond_click(O);
      
      return(1);
    }
  else
    {
      int tmp;

      tmp=0;
      highlight=0;
      
      if(lu) if(lu->respond_click(O)) tmp=1;
      if(ll) if(ll->respond_click(O)) tmp=1;
      if(ru) if(ru->respond_click(O)) tmp=1;
      if(rl) if(rl->respond_click(O)) tmp=1;

      return(tmp);
    }
}

//////////////////////////////////////////////////////////

QuadNode::QuadNode(Quadrilateral *Q)
{
  parent=NULL;
  next  =NULL;

  lu  = NULL;
  ll  = NULL;
  ru  = NULL;
  rl  = NULL;
  L   = NULL;
  level=0;

  QN   = NULL;
  QE   = NULL;
  QW   = NULL;
  QS   = NULL;

  if(Q) Quad=Q;
  else
    {
      printf("Error: Quadrilateral is NULL!\n");
      printf("[QuadNode::QuadNode(Quadrilateral *Q)]\n");
      exit(1);
    }
  selected=1;

  x =0.0;
  y =0.0;
  dx=1.0;
  dy=1.0;
  area=(Q->dx.Length())*(Q->dy.Length());
  E   =Q->E();
  rho =Q->rho();

  sub_divide();
}

//////////////////////////////////////////////////////////

void QuadNode::update_neighbor(QuadNode * N, char dir)
{


}

//////////////////////////////////////////////////////////

void QuadNode::set_neighbor(QuadNode *N, char dir)
{


}

//////////////////////////////////////////////////////////

Color QuadNode::get_Color(char dir)
{
  Color C;

  return(Bs);

  if(dir=='N') // northeast
    {
      // case 1:
      // four neighbors
      if(QN&&QE)
	{
	  C=0.25*(Bs+QN->Bs+QE->Bs+QE->QN->Bs);
	  return(C);
	}
      // case 3:
      // very northeast corner
      if(!(QN||QE))
	{
	  C=0.25*(7.0*Bs-QW->Bs-QS->Bs-QS->QW->Bs);
	  return(C);
	}
      // case 2:
      // north or east edge
      if(QN)
	{
	  C=0.25*(3.0*QN->Bs+3.0*Bs-QW->Bs-QW->QN->Bs);
	  return(C);
	}

      if(QE)
	{
	  C=0.25*(3.0*Bs+3.0*QE->Bs-QS->Bs-QS->QE->Bs);
	  return(C);
	}
    }
return(Bs);
  if(dir=='E') // southeast
    {

    }
  if(dir=='S') // southwest
    {

    }
  if(dir=='W') // northwest
    {

    }
}

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

void QuadTree::print_info(void)
{
  output(80,100,"QuadTree hit");
  printf("QuadTree hit\n");
}

//////////////////////////////////////////////////////////

void QuadTree::OpenGL_draw(void)
{
  if(root)
    {
      root->OpenGL_draw();
    }
}

//////////////////////////////////////////////////////////

int QuadTree::respond_click(int O)
{
  if(root)
    {
      //return(root->respond_click(O));

      if(root->respond_click(O))
	{
	  printf("*\n");
	  //root->OpenGL_draw();
	      if(root->L)
		{
		  root->L->OpenGL_draw();
		}
	      return(1);
	}
      
      if(root->L)
	{
	  if(root->L->respond_click(O))
	    {
	      root->L->OpenGL_draw();
	      return(1);
	    }
	}
    }
  return(0);
}

//////////////////////////////////////////////////////////
