//////////////////////////////////////////////////////////
// QuadTree.h, 2/14/2000, Ilja Friedel, cs 174
//
// class QuadTree
//////////////////////////////////////////////////////////

#ifndef QUADTREE_USED
#define QUADTREE_USED

#include <stdio.h>
#include <math.h>
#include "Vector.h"
#include "Object.h"

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

class LinkNode;

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

class QuadNode: public OpenGLobject
{

public:

  Color      Bg;
  Color      Bs;
  Color      E;
  double     area;
  Color      rho;

  LinkNode * L;
  QuadNode * next; // QuadNodeList

  QuadNode * QN; // for height ballancing
  QuadNode * QW; 
  QuadNode * QE;
  QuadNode * QS;

  QuadNode * parent;
  QuadNode * lu; // children
  QuadNode * ll; 
  QuadNode * ru;
  QuadNode * rl;

  int    level;

  double x,dx;
  double y,dy;

  Quadrilateral * Quad;

  int selected;

//////////////////////////////////////////////////////////

  virtual void OpenGL_draw(void);
  virtual void print_info(void);
  virtual int  respond_click(int O);
  virtual int  OGLID(void);

//////////////////////////////////////////////////////////

  Vector get_point(double a, double b);

  int    leaf(void)
    {
      if(lu&&ll&&ru&&rl) return(0);
      else return(1);
    }

//////////////////////////////////////////////////////////

  void highlight_branch(void);

  QuadNode(Quadrilateral *Q);
  QuadNode(QuadNode *p, int sector)
    {
      parent=p;
      lu  = NULL;
      ll  = NULL;
      ru  = NULL;
      rl  = NULL;
      L   = NULL;
 
      if(!p)
	{
	  printf("Error: Use other constructor for root!\n");
	  printf("[QuadNode(QuadNode *p, int sector)]\n");
	  exit(1);
	}
      
      level= p->level+1;
      area = 0.25*p->area;
      E    = p->E;
      rho  = p->rho;

      if(p->Quad) Quad= p->Quad;
      else
	{
	  printf("Error: Quadrilateral is NULL!\n");
	  printf("[QuadNode::QuadNode(QuadNode *p,int sector)]\n");
	  exit(1);
	}

      if(sector<0 || sector >=4)
	{
	  printf("Error: sector not in {0,1,2,3}!\n");
	  printf("[QuadNode(QuadNode *p, int sector)]\n");
	  exit(1);
	}

      dx=p->dx/2.0;
      dy=p->dy/2.0;

      if(sector==0)
	{
	  x =p->x+dx;
	  y =p->y+dy;
	  
	  QN   = p->QN;
	  QE   = p->QE;
	  QW   = p->lu;
	  QS   = p->rl;
	}

      if(sector==1)
	{
	  x =p->x;
	  y =p->y+dy;

	  QN   = p->QN;
	  QE   = p->ru;
	  QW   = p->QW;
	  QS   = p->ll;
	}

      if(sector==2)
	{
	  x =p->x;
	  y =p->y;

	  QN   = p->lu;
	  QE   = p->rl;
	  QW   = p->QW;
	  QS   = p->QS;
	}

      if(sector==3)
	{
	  x =p->x+dx;
	  y =p->y;

	  QN   = p->ru;
	  QE   = p->QE;
	  QW   = p->ll;
	  QS   = p->QS;
	}

      selected=1;

    }

//////////////////////////////////////////////////////////

  void sub_divide(void)
    {
      selected=0;
      
      if(!ru) ru=new QuadNode(this,0);
      if(!lu) lu=new QuadNode(this,1);
      if(!ll) ll=new QuadNode(this,2);
      if(!rl) rl=new QuadNode(this,3);

      lu->QN=QN;
      ru->QN=QN;
      ru->QE=QE;
      rl->QE=QE;
      rl->QS=QS;
      ll->QS=QS;
      ll->QW=QW;
      lu->QW=QW;

      lu->QE=ru;
      ru->QW=lu;
      ll->QE=rl;
      rl->QW=ll;

      lu->QS=ll;
      ll->QN=lu;
      ru->QS=rl;
      rl->QN=ru;
      
      rl->next=next;
      ll->next=rl;
      lu->next=ll;
      ru->next=lu;
      next=ru;

      if(QN) QN->update_neighbor(this,'S');
      if(QE) QE->update_neighbor(this,'W');
      if(QW) QW->update_neighbor(this,'E');
      if(QS) QS->update_neighbor(this,'N');
    }

//////////////////////////////////////////////////////////

  void  update_neighbor(QuadNode *N, char dir);
  void  set_neighbor(   QuadNode *N, char dir); 

  Color get_Color(char dir); 
};

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

class QuadTree : public OpenGLobject
{
public:

  QuadNode * root;

  QuadTree(Quadrilateral *Quad)
    {
      root=new QuadNode(Quad);
    }

  void highlight_branch(QuadNode *QN);

  virtual void print_info(void);
  virtual void OpenGL_draw(void);
  virtual int  respond_click(int O);

};

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

class LinkNode: public OpenGLobject
{
public:

  QuadNode * q; // shooter
  QuadNode * p; // receiver
  double     Fpq;
  LinkNode * next;


  LinkNode(QuadNode *s, QuadNode *r, double F)
    {
      p=s;
      q=r;
      Fpq=F;
      next=q->L;
      q->L=this;
    }

  ~LinkNode()
    {
      if(q->L==this)
	{
	  q->L=next;
	}
      else
	{
	  LinkNode *l;
	  
	  l=q->L;
	  while(l->next!=this)
	    {
	      l=l->next;
	    }
	  l->next=next;	  
	}
    }

  void Ballance(void);
 
  virtual void OpenGL_draw(void);
  virtual void print_info(void);
  virtual int  respond_click(int O);
  virtual int  OGLID(void);
};

//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////

#endif
